#!/usr/bin/env python3
"""
Wireless Sniffer
Wireless packet capture and analysis tool.
Parameters:
    - interface: Wireless interface to use
    - filter: BPF filter for packet capture
    - output: PCAP file to save captured packets
Requires: scapy package and admin/root privileges
"""
from scapy.all import *
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import time
import os
import signal
import binascii

class WirelessSniffer:
    def __init__(self, interface, packet_filter=None, pcap_file=None):
        self.interface = interface
        self.packet_filter = packet_filter
        self.pcap_file = pcap_file
        self.stop_event = threading.Event()
        self.packets = []
        self.stats = defaultdict(int)
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def packet_handler(self, pkt):
        """Process captured packets"""
        if self.stop_event.is_set():
            return

        try:
            packet_info = {
                'timestamp': datetime.now().isoformat(),
                'size': len(pkt),
                'summary': pkt.summary()
            }

            # Update statistics
            self.stats['total_packets'] += 1
            self.stats['total_bytes'] += packet_info['size']

            # Layer 2 analysis
            if pkt.haslayer(Dot11):
                packet_info['layer2'] = {
                    'type': pkt[Dot11].type,
                    'subtype': pkt[Dot11].subtype
                }
                self.stats['802.11_packets'] += 1

                # Management frames
                if pkt[Dot11].type == 0:
                    self.stats['management_frames'] += 1
                    if pkt.haslayer(Dot11Beacon):
                        self.stats['beacons'] += 1
                        packet_info['beacon'] = {
                            'ssid': pkt[Dot11Elt].info.decode('utf-8',
                                                            errors='ignore'),
                            'channel': int(ord(pkt[Dot11Elt:3].info))
                        }
                    elif pkt.haslayer(Dot11ProbeReq):
                        self.stats['probe_requests'] += 1
                    elif pkt.haslayer(Dot11ProbeResp):
                        self.stats['probe_responses'] += 1
                    elif pkt.haslayer(Dot11Auth):
                        self.stats['auth_frames'] += 1
                    elif pkt.haslayer(Dot11AssoReq):
                        self.stats['assoc_requests'] += 1
                    elif pkt.haslayer(Dot11AssoResp):
                        self.stats['assoc_responses'] += 1

                # Control frames
                elif pkt[Dot11].type == 1:
                    self.stats['control_frames'] += 1
                    if pkt.haslayer(Dot11RTS):
                        self.stats['rts_frames'] += 1
                    elif pkt.haslayer(Dot11CTS):
                        self.stats['cts_frames'] += 1
                    elif pkt.haslayer(Dot11ACK):
                        self.stats['ack_frames'] += 1

                # Data frames
                elif pkt[Dot11].type == 2:
                    self.stats['data_frames'] += 1
                    if pkt.haslayer(Dot11WEP):
                        self.stats['encrypted_frames'] += 1
                        packet_info['encryption'] = 'WEP'
                    elif pkt.haslayer(Dot11CCMP):
                        self.stats['encrypted_frames'] += 1
                        packet_info['encryption'] = 'CCMP (AES)'
                    elif pkt.haslayer(Dot11TKIP):
                        self.stats['encrypted_frames'] += 1
                        packet_info['encryption'] = 'TKIP'

            # Layer 3 analysis
            if pkt.haslayer(IP):
                packet_info['layer3'] = {
                    'src': pkt[IP].src,
                    'dst': pkt[IP].dst,
                    'proto': pkt[IP].proto
                }
                self.stats['ip_packets'] += 1

                # TCP analysis
                if pkt.haslayer(TCP):
                    packet_info['tcp'] = {
                        'sport': pkt[TCP].sport,
                        'dport': pkt[TCP].dport,
                        'flags': pkt[TCP].flags
                    }
                    self.stats['tcp_packets'] += 1

                # UDP analysis
                elif pkt.haslayer(UDP):
                    packet_info['udp'] = {
                        'sport': pkt[UDP].sport,
                        'dport': pkt[UDP].dport
                    }
                    self.stats['udp_packets'] += 1

                # ICMP analysis
                elif pkt.haslayer(ICMP):
                    packet_info['icmp'] = {
                        'type': pkt[ICMP].type,
                        'code': pkt[ICMP].code
                    }
                    self.stats['icmp_packets'] += 1

            # Store packet
            if self.pcap_file:
                wrpcap(self.pcap_file, pkt, append=True)
            self.packets.append(packet_info)

        except Exception as e:
            self.logger.error(f"Error processing packet: {str(e)}")

    def analyze_traffic_patterns(self):
        """Analyze captured traffic for patterns"""
        analysis = {
            'frame_distribution': {
                'management': self.stats['management_frames'],
                'control': self.stats['control_frames'],
                'data': self.stats['data_frames']
            },
            'encryption_stats': {
                'encrypted': self.stats['encrypted_frames'],
                'unencrypted': (self.stats['data_frames'] -
                              self.stats['encrypted_frames'])
            },
            'protocol_distribution': {
                'tcp': self.stats['tcp_packets'],
                'udp': self.stats['udp_packets'],
                'icmp': self.stats['icmp_packets']
            }
        }
        return analysis

    def save_results(self, output_file):
        results = {
            'timestamp': datetime.now().isoformat(),
            'interface': self.interface,
            'filter': self.packet_filter,
            'statistics': dict(self.stats),
            'analysis': self.analyze_traffic_patterns(),
            'packets': self.packets[:1000]  # Limit to prevent huge files
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_stats(self):
        """Print current capture statistics"""
        self.logger.info("\nCapture Statistics:")
        self.logger.info(f"Total Packets: {self.stats['total_packets']}")
        self.logger.info(f"Total Bytes: {self.stats['total_bytes']}")
        
        # Frame type distribution
        self.logger.info("\nFrame Type Distribution:")
        self.logger.info(f"Management Frames: {self.stats['management_frames']}")
        self.logger.info(f"Control Frames: {self.stats['control_frames']}")
        self.logger.info(f"Data Frames: {self.stats['data_frames']}")
        
        # Management frame details
        self.logger.info("\nManagement Frame Details:")
        self.logger.info(f"Beacons: {self.stats['beacons']}")
        self.logger.info(f"Probe Requests: {self.stats['probe_requests']}")
        self.logger.info(f"Probe Responses: {self.stats['probe_responses']}")
        self.logger.info(f"Authentication: {self.stats['auth_frames']}")
        
        # Data frame details
        self.logger.info("\nData Frame Details:")
        self.logger.info(f"Encrypted Frames: {self.stats['encrypted_frames']}")
        
        # Protocol distribution
        self.logger.info("\nProtocol Distribution:")
        self.logger.info(f"TCP Packets: {self.stats['tcp_packets']}")
        self.logger.info(f"UDP Packets: {self.stats['udp_packets']}")
        self.logger.info(f"ICMP Packets: {self.stats['icmp_packets']}")

    def start_capture(self, output_file=None):
        try:
            self.logger.info(f"Starting packet capture on {self.interface}")
            if self.packet_filter:
                self.logger.info(f"Using filter: {self.packet_filter}")
            if self.pcap_file:
                self.logger.info(f"Saving packets to: {self.pcap_file}")

            # Start statistics printer thread
            def print_stats_periodically():
                while not self.stop_event.is_set():
                    self.print_stats()
                    time.sleep(5)

            stats_thread = threading.Thread(target=print_stats_periodically)
            stats_thread.daemon = True
            stats_thread.start()

            # Start packet capture
            sniff(
                iface=self.interface,
                prn=self.packet_handler,
                filter=self.packet_filter,
                store=0,
                stop_filter=lambda _: self.stop_event.is_set()
            )

        except KeyboardInterrupt:
            self.logger.info("Capture interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during capture: {str(e)}")
        finally:
            self.stop_event.set()
            if output_file:
                self.save_results(output_file)
            self.print_stats()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Wireless Sniffer")
    parser.add_argument("interface", help="Wireless interface to use")
    parser.add_argument("--filter", help="BPF filter for packet capture")
    parser.add_argument("--pcap", help="PCAP file to save captured packets")
    parser.add_argument("--output", help="Output file for analysis results")
    args = parser.parse_args()

    sniffer = WirelessSniffer(args.interface, args.filter, args.pcap)
    sniffer.start_capture(args.output)
